part of gauges;

/// Represents the container to render the axis and its element
///
class _AxisContainer extends StatelessWidget {
  const _AxisContainer(
      this._gauge, this._gaugeThemeData, this._animationDetails);

  /// Hold the radial gauge animation details
  final _AnimationDetails _animationDetails;

  /// Specifies the gauge theme data.
  final SfGaugeThemeData _gaugeThemeData;

  /// Specifies the radial gauge
  final SfRadialGauge _gauge;

  /// Method to update the pointer value
  void _updatePointerValue(BuildContext context, DragUpdateDetails details) {
    final RenderBox renderBox = context.findRenderObject();
    if (details != null && details.globalPosition != null) {
      final Offset tapPosition =
          renderBox.globalToLocal(details.globalPosition);
      for (num i = 0; i < _gauge.axes.length; i++) {
        if (_gauge.axes[i].pointers != null &&
            _gauge.axes[i].pointers.isNotEmpty) {
          for (num j = 0; j < _gauge.axes[i].pointers.length; j++) {
            final GaugePointer pointer = _gauge.axes[i].pointers[j];
            if (pointer.enableDragging && pointer._isDragStarted) {
              final Rect rect = Rect.fromLTRB(
                  pointer._axis._axisRect.left + pointer._axis._axisCenter.dx,
                  pointer._axis._axisRect.top + pointer._axis._axisCenter.dy,
                  pointer._axis._axisRect.right + pointer._axis._axisCenter.dx,
                  pointer._axis._axisRect.bottom +
                      pointer._axis._axisCenter.dy);
              if (pointer is RangePointer) {
                final RangePointer rangePointer = pointer;

                final double actualCenterX = rangePointer._pointerRect.left +
                    rangePointer._axis._axisCenter.dx +
                    rangePointer._axis._radius;
                final double actualCenterY = rangePointer._pointerRect.top +
                    rangePointer._axis._axisCenter.dy +
                    rangePointer._axis._radius;
                final double x = tapPosition.dx - actualCenterX;
                final double y = tapPosition.dy - actualCenterY;

                /// Checks whether the tapped position is available inside the
                /// radius of range pointer
                final bool isInside = (x * x) + (y * y) <=
                    (rangePointer._axis._radius * rangePointer._axis._radius);
                if (isInside) {
                  pointer._updateDragValue(
                      tapPosition.dx, tapPosition.dy, _animationDetails);
                }
              }

              /// Checks whether the tapped position is available inside the
              /// rect of needle or marker pointer
              else if (rect.contains(tapPosition)) {
                pointer._updateDragValue(
                    tapPosition.dx, tapPosition.dy, _animationDetails);
              }
            }
          }
        }
      }
    }
  }

  /// To initialize the gauge elements
  Widget _getGaugeElements(BuildContext context, BoxConstraints constraints) {
    final List<Widget> gaugeWidgets = <Widget>[];
    _calculateAxisElementPosition(context, constraints);
    _addGaugeElements(constraints, context, gaugeWidgets);
    final bool enableAxisTapping = _getIsAxisTapped();
    final bool enablePointerDragging = _getIsPointerDragging();

    return GestureDetector(
        onPanStart: enablePointerDragging
            ? (DragStartDetails details) =>
                _checkPointerDragging(context, details)
            : null,
        onPanUpdate: enablePointerDragging
            ? (DragUpdateDetails details) =>
                _updatePointerValue(context, details)
            : null,
        onPanEnd: enablePointerDragging
            ? (DragEndDetails details) => _checkPointerIsDragged()
            : null,
        onTapUp: enableAxisTapping
            ? (TapUpDetails details) => _checkIsAxisTapped(context, details)
            : null,
        child: Container(
          decoration: const BoxDecoration(color: Colors.transparent),
          child: Stack(children: gaugeWidgets),
        ));
  }

  // Checks whether onAxisTapped callback is activated.
  bool _getIsAxisTapped() {
    for (int i = 0; i < _gauge.axes.length; i++) {
      if (_gauge.axes[i].onAxisTapped != null) {
        return true;
      }
    }
    return false;
  }

  // Checks whether the pointer dragging is enabled
  bool _getIsPointerDragging() {
    for (int i = 0; i < _gauge.axes.length; i++) {
      final RadialAxis currentAxis = _gauge.axes[i];
      if (currentAxis.pointers != null && currentAxis.pointers.isNotEmpty) {
        for (int j = 0; j < currentAxis.pointers.length; j++) {
          if (currentAxis.pointers[j].enableDragging) {
            return true;
          }
        }
      }
    }
    return false;
  }

  /// Method to check whether the axis is tapped
  void _checkIsAxisTapped(BuildContext context, TapUpDetails details) {
    final RenderBox renderBox = context.findRenderObject();
    if (details != null && details.globalPosition != null) {
      if (_gauge.axes.isNotEmpty) {
        for (num i = 0; i < _gauge.axes.length; i++) {
          final RadialAxis axis = _gauge.axes[i];
          final Offset offset = renderBox.globalToLocal(details.globalPosition);
          if (axis.onAxisTapped != null &&
              axis._axisPath.getBounds().contains(offset)) {
            axis._calculateAngleFromOffset(offset);
          }
        }
      }
    }
  }

  /// Method to check whether the axis pointer is dragging
  void _checkPointerDragging(BuildContext context, DragStartDetails details) {
    final RenderBox renderBox = context.findRenderObject();
    if (details != null && details.globalPosition != null) {
      final Offset tapPosition =
          renderBox.globalToLocal(details.globalPosition);
      for (num i = 0; i < _gauge.axes.length; i++) {
        if (_gauge.axes[i].pointers != null &&
            _gauge.axes[i].pointers.isNotEmpty) {
          for (num j = 0; j < _gauge.axes[i].pointers.length; j++) {
            final GaugePointer pointer = _gauge.axes[i].pointers[j];
            if (pointer.enableDragging) {
              if (pointer is RangePointer) {
                final RangePointer rangePointer = pointer;
                final RadialAxis axis = _gauge.axes[i];
                final Rect pathRect = Rect.fromLTRB(
                    rangePointer._arcPath.getBounds().left +
                        axis._axisCenter.dx -
                        5,
                    rangePointer._arcPath.getBounds().top +
                        axis._axisCenter.dy -
                        5,
                    rangePointer._arcPath.getBounds().right +
                        axis._axisCenter.dx +
                        5,
                    rangePointer._arcPath.getBounds().bottom +
                        axis._axisCenter.dy +
                        5);

                // Checks whether the tapped position is present inside
                // the range pointer path
                if (pathRect.contains(tapPosition)) {
                  rangePointer._isDragStarted = true;
                  pointer._createPointerValueChangeStartArgs();
                  break;
                } else {
                  rangePointer._isDragStarted = false;
                }
              } else {
                if (pointer._pointerRect.contains(tapPosition)) {
                  pointer._isDragStarted = true;
                  pointer._createPointerValueChangeStartArgs();
                  break;
                } else {
                  pointer._isDragStarted = false;
                }
              }
            }
          }
        }
      }
    }
  }

  /// Method to ensure whether the pointer was dragged
  void _checkPointerIsDragged() {
    if (_gauge.axes != null) {
      for (num i = 0; i < _gauge.axes.length; i++) {
        if (_gauge.axes[i].pointers != null &&
            _gauge.axes[i].pointers.isNotEmpty) {
          for (num j = 0; j < _gauge.axes[i].pointers.length; j++) {
            final GaugePointer pointer = _gauge.axes[i].pointers[j];
            if (pointer.enableDragging) {
              if (pointer._isDragStarted) {
                pointer._createPointerValueChangeEndArgs();

                if (pointer is NeedlePointer) {
                  final NeedlePointer needlePointer = pointer;
                  needlePointer._animationEndValue =
                      needlePointer._getSweepAngle();
                } else if (pointer is MarkerPointer) {
                  final MarkerPointer markerPointer = pointer;
                  markerPointer._animationEndValue =
                      markerPointer._getSweepAngle();
                } else {
                  final RangePointer rangePointer = pointer;
                  rangePointer._animationEndValue =
                      rangePointer._getSweepAngle();
                }
              }
              pointer._isDragStarted = false;
            }
          }
        }
      }
    }
  }

  /// Calculates the axis position
  void _calculateAxisElementPosition(
      BuildContext context, BoxConstraints constraints) {
    if (_gauge.axes != null && _gauge.axes.isNotEmpty) {
      for (int i = 0; i < _gauge.axes.length; i++) {
        _gauge.axes[i]._calculateAxisRange(
            constraints, context, _gaugeThemeData, _animationDetails);
      }
    }
  }

  /// Methods to add the gauge elements
  void _addGaugeElements(BoxConstraints constraints, BuildContext context,
      List<Widget> gaugeWidgets) {
    if (_gauge.axes != null && _gauge.axes.isNotEmpty) {
      final _AnimationDurationDetails durationDetails =
          _AnimationDurationDetails();
      _calculateDurationForAnimation(durationDetails);
      for (int i = 0; i < _gauge.axes.length; i++) {
        final RadialAxis axis = _gauge.axes[i];
        _addAxis(axis, constraints, gaugeWidgets, durationDetails);

        if (axis.ranges != null && axis.ranges.isNotEmpty) {
          _addRange(axis, constraints, gaugeWidgets, durationDetails);
        }

        if (axis.pointers != null && axis.pointers.isNotEmpty) {
          for (num j = 0; j < axis.pointers.length; j++) {
            if (axis.pointers[j] is NeedlePointer) {
              final NeedlePointer needlePointer = axis.pointers[j];
              _addNeedlePointer(axis, constraints, needlePointer, gaugeWidgets,
                  durationDetails);
            } else if (axis.pointers[j] is MarkerPointer) {
              final MarkerPointer markerPointer = axis.pointers[j];
              _addMarkerPointer(axis, constraints, markerPointer, gaugeWidgets,
                  durationDetails);
            } else if (axis.pointers[j] is RangePointer) {
              final RangePointer rangePointer = axis.pointers[j];
              _addRangePointer(axis, constraints, rangePointer, gaugeWidgets,
                  durationDetails);
            }
          }
        }

        if (axis.annotations != null && axis.annotations.isNotEmpty) {
          _addAnnotation(axis, gaugeWidgets, durationDetails);
        }
      }
    }
  }

  /// Adds the axis
  void _addAxis(RadialAxis axis, BoxConstraints constraints,
      List<Widget> gaugeWidgets, _AnimationDurationDetails durationDetails) {
    Animation<double> axisAnimation;
    Animation<double> axisElementAnimation;
    if (_animationDetails.needsToAnimateAxes &&
        (durationDetails.hasAxisLine || durationDetails.hasAxisElements)) {
      _animationDetails.animationController.duration =
          Duration(milliseconds: _gauge.animationDuration.toInt());
      // Includes animation duration for axis line
      if (durationDetails.hasAxisLine) {
        axisAnimation = Tween<double>(begin: 0, end: 1).animate(CurvedAnimation(
            parent: _animationDetails.animationController,
            curve: Interval(durationDetails.axisLineInterval[0],
                durationDetails.axisLineInterval[1],
                curve: Curves.easeIn)));
      }
      // Includes animation duration for axis ticks and labels
      if (durationDetails.hasAxisElements) {
        axisElementAnimation = Tween<double>(begin: 0, end: 1).animate(
            CurvedAnimation(
                parent: _animationDetails.animationController,
                curve: Interval(durationDetails.axisElementsInterval[0],
                    durationDetails.axisElementsInterval[1],
                    curve: Curves.easeIn)));
      }
    }

    gaugeWidgets.add(Container(
      child: RepaintBoundary(
        child: CustomPaint(
            painter: _AxisPainter(
                _gauge,
                axis,
                axis._needsRepaintAxis ?? true,
                _animationDetails.axisRepaintNotifier,
                axisAnimation,
                axisElementAnimation,
                _gaugeThemeData,
                _animationDetails),
            size: Size(constraints.maxWidth, constraints.maxHeight)),
      ),
    ));

    if (axisAnimation != null || axisElementAnimation != null) {
      _animationDetails.animationController.forward(from: 0.0);
    }
  }

  /// Adds the range pointer
  void _addRangePointer(
      RadialAxis axis,
      BoxConstraints constraints,
      RangePointer rangePointer,
      List<Widget> gaugeWidgets,
      _AnimationDurationDetails durationDetails) {
    rangePointer._animationEndValue = rangePointer._getSweepAngle();
    Animation<double> pointerAnimation;
    final List<double> animationIntervals =
        _getPointerAnimationInterval(durationDetails);
    if (_animationDetails.needsToAnimatePointers ||
        (rangePointer.enableAnimation &&
            rangePointer.animationDuration > 0 &&
            rangePointer._needsAnimate)) {
      _animationDetails.animationController.duration = Duration(
          milliseconds:
              _getPointerAnimationDuration(rangePointer.animationDuration));
      final Curve pointerAnimationType =
          _getCurveAnimation(rangePointer.animationType);
      final double endValue = rangePointer._animationEndValue;
      pointerAnimation = Tween<double>(
              begin: rangePointer._animationStartValue ?? 0, end: endValue)
          .animate(CurvedAnimation(
              parent: _animationDetails.animationController,
              curve: Interval(animationIntervals[0], animationIntervals[1],
                  curve: pointerAnimationType)));
    }

    gaugeWidgets.add(Container(
      child: RepaintBoundary(
        child: CustomPaint(
            painter: _RangePointerPainter(
                _gauge,
                axis,
                rangePointer,
                rangePointer._needsRepaintPointer ?? true,
                pointerAnimation,
                _animationDetails.pointerRepaintNotifier,
                _gaugeThemeData,
                _animationDetails),
            size: Size(constraints.maxWidth, constraints.maxHeight)),
      ),
    ));
    if (_animationDetails.needsToAnimatePointers ||
        rangePointer._getIsPointerAnimationEnabled()) {
      _animationDetails.animationController.forward(from: 0.0);
    }
  }

  /// Adds the needle pointer
  void _addNeedlePointer(
      RadialAxis axis,
      BoxConstraints constraints,
      NeedlePointer needlePointer,
      List<Widget> gaugeWidgets,
      _AnimationDurationDetails durationDetails) {
    Animation<double> pointerAnimation;
    final List<double> intervals =
        _getPointerAnimationInterval(durationDetails);
    needlePointer._animationEndValue = needlePointer._getSweepAngle();
    if (_animationDetails.needsToAnimatePointers ||
        (needlePointer.enableAnimation &&
            needlePointer.animationDuration > 0 &&
            needlePointer._needsAnimate)) {
      _animationDetails.animationController.duration = Duration(
          milliseconds:
              _getPointerAnimationDuration(needlePointer.animationDuration));

      final double startValue = axis.isInversed ? 1 : 0;
      final double endValue = axis.isInversed ? 0 : 1;
      double actualValue = needlePointer._animationStartValue ?? startValue;
      actualValue = actualValue == endValue ? startValue : actualValue;
      pointerAnimation = Tween<double>(
              begin: actualValue, end: needlePointer._animationEndValue)
          .animate(CurvedAnimation(
              parent: _animationDetails.animationController,
              curve: Interval(intervals[0], intervals[1],
                  curve: _getCurveAnimation(needlePointer.animationType))));
    }

    gaugeWidgets.add(Container(
      child: RepaintBoundary(
        child: CustomPaint(
            painter: _NeedlePointerPainter(
                _gauge,
                axis,
                needlePointer,
                needlePointer._needsRepaintPointer ?? true,
                pointerAnimation,
                _animationDetails.pointerRepaintNotifier,
                _gaugeThemeData,
                _animationDetails),
            size: Size(constraints.maxWidth, constraints.maxHeight)),
      ),
    ));

    if (_animationDetails.needsToAnimatePointers ||
        needlePointer._getIsPointerAnimationEnabled()) {
      _animationDetails.animationController.forward(from: 0.0);
    }
  }

  /// Adds the marker pointer
  void _addMarkerPointer(
      RadialAxis axis,
      BoxConstraints constraints,
      MarkerPointer markerPointer,
      List<Widget> gaugeWidgets,
      _AnimationDurationDetails durationDetails) {
    Animation<double> pointerAnimation;
    final List<double> pointerIntervals =
        _getPointerAnimationInterval(durationDetails);
    markerPointer._animationEndValue = markerPointer._getSweepAngle();
    if (_animationDetails.needsToAnimatePointers ||
        (markerPointer.enableAnimation &&
            markerPointer.animationDuration > 0 &&
            markerPointer._needsAnimate)) {
      _animationDetails.animationController.duration = Duration(
          milliseconds:
              _getPointerAnimationDuration(markerPointer.animationDuration));

      final double startValue = axis.isInversed ? 1 : 0;
      pointerAnimation = Tween<double>(
              begin: markerPointer._animationStartValue ?? startValue,
              end: markerPointer._animationEndValue)
          .animate(CurvedAnimation(
              parent: _animationDetails.animationController,
              curve: Interval(pointerIntervals[0], pointerIntervals[1],
                  curve: _getCurveAnimation(markerPointer.animationType))));
    }

    gaugeWidgets.add(Container(
      child: RepaintBoundary(
        child: CustomPaint(
            painter: _MarkerPointerPainter(
                _gauge,
                axis,
                markerPointer,
                markerPointer._needsRepaintPointer ?? true,
                pointerAnimation,
                _animationDetails.pointerRepaintNotifier,
                _gaugeThemeData,
                _animationDetails),
            size: Size(constraints.maxWidth, constraints.maxHeight)),
      ),
    ));

    if (_animationDetails.needsToAnimatePointers ||
        markerPointer._getIsPointerAnimationEnabled()) {
      _animationDetails.animationController.forward(from: 0.0);
    }
  }

  /// Adds the axis range
  void _addRange(RadialAxis axis, BoxConstraints constraints,
      List<Widget> gaugeWidgets, _AnimationDurationDetails durationDetails) {
    for (num k = 0; k < axis.ranges.length; k++) {
      Animation<double> rangeAnimation;
      if (_animationDetails.needsToAnimateRanges) {
        _animationDetails.animationController.duration =
            Duration(milliseconds: _gauge.animationDuration.toInt());
        rangeAnimation = Tween<double>(begin: 0, end: 1).animate(
            CurvedAnimation(
                parent: _animationDetails.animationController,
                curve: Interval(durationDetails.rangesInterval[0],
                    durationDetails.rangesInterval[1],
                    curve: Curves.easeIn)));
      }

      gaugeWidgets.add(Container(
        child: RepaintBoundary(
          child: CustomPaint(
              painter: _RangePainter(
                  _gauge,
                  axis,
                  axis.ranges[k],
                  axis.ranges[k]._needsRepaintRange ?? true,
                  rangeAnimation,
                  _animationDetails.rangeRepaintNotifier,
                  _gaugeThemeData,
                  _animationDetails),
              size: Size(constraints.maxWidth, constraints.maxHeight)),
        ),
      ));

      if (rangeAnimation != null) {
        _animationDetails.animationController.forward(from: 0.0);
      }
    }
  }

  /// Return the animation duration
  int _getPointerAnimationDuration(double duration) {
    if (_animationDetails.needsToAnimatePointers) {
      return _gauge.animationDuration.toInt();
    } else {
      return duration.toInt();
    }
  }

  /// Returns the animation interval of pointers
  List<double> _getPointerAnimationInterval(
      _AnimationDurationDetails durationDetails) {
    List<double> pointerIntervals = List<double>(2);
    if (_animationDetails.needsToAnimatePointers) {
      pointerIntervals = durationDetails.pointersInterval;
    } else {
      pointerIntervals[0] = 0;
      pointerIntervals[1] = 1;
    }

    return pointerIntervals;
  }

  /// Adds the axis annotation
  void _addAnnotation(RadialAxis axis, List<Widget> gaugeWidgets,
      _AnimationDurationDetails durationDetails) {
    for (num j = 0; j < axis.annotations.length; j++) {
      final GaugeAnnotation annotation = axis.annotations[j];
      gaugeWidgets.add(_AnnotationRenderer(
        annotation: annotation,
        gauge: _gauge,
        axis: axis,
        interval: durationDetails.annotationInterval,
        duration: _gauge.animationDuration.toInt(),
        animationDetails: _animationDetails,
      ));
    }
  }

  ///calculates the duration for animation
  void _calculateDurationForAnimation(
      _AnimationDurationDetails durationDetails) {
    num totalCount = 5;
    double interval;
    double startValue = 0.05;
    double endValue = 0;
    for (num i = 0; i < _gauge.axes.length; i++) {
      _calculateAxisElements(_gauge.axes[i], durationDetails);
    }

    totalCount = _getElementsCount(totalCount, durationDetails);

    interval = 1 / totalCount;
    endValue = interval;
    if (durationDetails.hasAxisLine) {
      durationDetails.axisLineInterval = List<double>(2);
      durationDetails.axisLineInterval[0] = startValue;
      durationDetails.axisLineInterval[1] = endValue;
      startValue = endValue;
      endValue += interval;
    }

    if (durationDetails.hasAxisElements) {
      durationDetails.axisElementsInterval = List<double>(2);
      durationDetails.axisElementsInterval[0] = startValue;
      durationDetails.axisElementsInterval[1] = endValue;
      startValue = endValue;
      endValue += interval;
    }

    if (durationDetails.hasRanges) {
      durationDetails.rangesInterval = List<double>(2);
      durationDetails.rangesInterval[0] = startValue;
      durationDetails.rangesInterval[1] = endValue;
      startValue = endValue;
      endValue += interval;
    }

    if (durationDetails.hasPointers) {
      durationDetails.pointersInterval = List<double>(2);
      durationDetails.pointersInterval[0] = startValue;
      durationDetails.pointersInterval[1] = endValue;
      startValue = endValue;
      endValue += interval;
    }

    if (durationDetails.hasAnnotations) {
      durationDetails.annotationInterval = List<double>(2);
      durationDetails.annotationInterval[0] = startValue;
      durationDetails.annotationInterval[1] = endValue;
    }
  }

  /// Returns the total elements count
  num _getElementsCount(
      num totalCount, _AnimationDurationDetails durationDetails) {
    if (!durationDetails.hasAnnotations) {
      totalCount -= 1;
    }

    if (!durationDetails.hasPointers) {
      totalCount -= 1;
    }

    if (!durationDetails.hasRanges) {
      totalCount -= 1;
    }

    if (!durationDetails.hasAxisElements) {
      totalCount -= 1;
    }

    if (!durationDetails.hasAxisLine) {
      totalCount -= 1;
    }

    return totalCount;
  }

  /// Calculates the  gauge elements
  void _calculateAxisElements(
      RadialAxis axis, _AnimationDurationDetails durationDetails) {
    if (axis.showAxisLine && !durationDetails.hasAxisLine) {
      durationDetails.hasAxisLine = true;
    }

    if ((axis.showTicks || axis.showLabels) &&
        !durationDetails.hasAxisElements) {
      durationDetails.hasAxisElements = true;
    }

    if (axis.ranges != null &&
        axis.ranges.isNotEmpty &&
        !durationDetails.hasRanges) {
      durationDetails.hasRanges = true;
    }

    if (axis.pointers != null &&
        axis.pointers.isNotEmpty &&
        !durationDetails.hasPointers) {
      durationDetails.hasPointers = true;
    }

    if (axis.annotations != null &&
        axis.annotations.isNotEmpty &&
        !durationDetails.hasAnnotations) {
      durationDetails.hasAnnotations = true;
    }
  }

  /// Method returns the curve animation function based on the animation type
  Curve _getCurveAnimation(AnimationType type) {
    Curve curve = Curves.linear;
    switch (type) {
      case AnimationType.bounceOut:
        curve = Curves.bounceOut;
        break;
      case AnimationType.ease:
        curve = Curves.ease;
        break;
      case AnimationType.easeInCirc:
        curve = Curves.easeInCirc;
        break;
      case AnimationType.easeOutBack:
        curve = Curves.easeOutBack;
        break;
      case AnimationType.elasticOut:
        curve = Curves.elasticOut;
        break;
      case AnimationType.linear:
        curve = Curves.linear;
        break;
      case AnimationType.slowMiddle:
        curve = Curves.slowMiddle;
        break;
    }
    return curve;
  }

  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(
        builder: (BuildContext context, BoxConstraints constraints) {
      return Container(
        decoration: const BoxDecoration(color: Colors.transparent),
        child: _getGaugeElements(context, constraints),
      );
    });
  }
}

/// Holds the load time animation duration details.
class _AnimationDurationDetails {
  /// Specifies the axis line interval for animation
  List<double> axisLineInterval;

  /// Specifies the axis element interval for load time animation
  List<double> axisElementsInterval;

  /// Specifies the range interval for initial animation
  List<double> rangesInterval;

  /// Specifies the pointer interval for load time animation
  List<double> pointersInterval;

  /// Specifies the annotation interval for load time animation
  List<double> annotationInterval;

  /// Specifies whether the axis line is enabled
  bool hasAxisLine = false;

  /// Specifies whether the axis element is enabled
  bool hasAxisElements = false;

  /// Specifies whether axis range is enabled
  bool hasRanges = false;

  /// Specifies whether the axis pointers is enabled
  bool hasPointers = false;

  /// Specifies whether the annotation is added
  bool hasAnnotations = false;
}
