part of gauges;

/// Represents the painter to draw marker
class _MarkerPointerPainter extends CustomPainter {
  _MarkerPointerPainter(
      this._gauge,
      this._axis,
      this._markerPointer,
      this._isRepaint,
      this._pointerAnimation,
      ValueNotifier<num> notifier,
      this._gaugeThemeData,
      this._animationDetails)
      : super(repaint: notifier);

  /// Specifies the circular gauge
  final SfRadialGauge _gauge;

  /// Specifies whether to repaint the series
  final bool _isRepaint;

  /// Specifies the axis of the painter
  final RadialAxis _axis;

  /// Specifies the marker pointer
  final MarkerPointer _markerPointer;

  /// Specifies the pointer animation
  final Animation<double> _pointerAnimation;

  /// Specifies the gauge theme data.
  final SfGaugeThemeData _gaugeThemeData;

  /// Hold the radial gauge animation details
  final _AnimationDetails _animationDetails;

  @override
  void paint(Canvas canvas, Size size) {
    final bool needsPointerAnimation =
        _markerPointer._getIsPointerAnimationEnabled();
    double markerAngle = 0;
    Offset offset;
    bool needsShowPointer;
    if (_pointerAnimation != null) {
      needsShowPointer = _axis.isInversed
          ? _pointerAnimation.value < 1
          : _pointerAnimation.value > 0;
    }

    if (_animationDetails.needsToAnimatePointers || needsPointerAnimation) {
      if ((_animationDetails.needsToAnimatePointers && needsShowPointer) ||
          !_animationDetails.needsToAnimatePointers) {
        markerAngle =
            (_axis._sweepAngle * _pointerAnimation.value) + _axis.startAngle;
        offset =
            _markerPointer._getMarkerOffset(_getDegreeToRadian(markerAngle));
      }
    } else {
      offset = _markerPointer._offset;
      markerAngle = _markerPointer._angle;
    }
    if (offset != null) {
      final PointerPaintingDetails pointerPaintingDetails =
          PointerPaintingDetails(
              startOffset: offset,
              endOffset: offset,
              pointerAngle: markerAngle,
              axisRadius: _axis._radius,
              axisCenter: _axis._axisCenter);
      _markerPointer.drawPointer(
          canvas, pointerPaintingDetails, _gaugeThemeData);
    }

    // Disables the animation once the animation reached the current
    // pointer angle
    if (needsPointerAnimation &&
        markerAngle ==
            _axis._sweepAngle * _markerPointer._animationEndValue +
                _axis.startAngle) {
      _markerPointer._needsAnimate = false;
    }

    if (_animationDetails.needsToAnimatePointers &&
        _gauge.axes[_gauge.axes.length - 1] == _axis &&
        _axis.pointers[_axis.pointers.length - 1] == _markerPointer &&
        markerAngle ==
            _axis._sweepAngle * _markerPointer._animationEndValue +
                _axis.startAngle) {
      _animationDetails.needsToAnimatePointers = false;
    }
  }

  @override
  bool shouldRepaint(_MarkerPointerPainter oldDelegate) => _isRepaint;
}
